/* !!! problem avec srand ou time!
   !!! bait f e 2 plante! (division par 0, apparement dans MT)
    __          __ __
   |  |__ _____|__|  |__   Original game "FootSteps" by Gordon Munroe
   |  _  |  _  |  |   __|  C version by Alain Brobecker (2004 august)
   |_____|___._|__|_____|

If you want to create your own computer player, you must first increment
nbplayers to include yours. Then you must write 4 routines which are:

1) void init() { ... }                     //called when bait program starts.
2) int bet(mypot,hispot,pointer,turn)      //called at every turn, it must
   int mypot[],hispot[],pointer[],turn; {  //  return a value between 1 and
     ...                                   //  mypot[turn] (0 is allowed
   }                                       //  only if mypot[turn]==0).
3) void endgame(mypot,hispot,pointer,turn) //called at the end of each game.
   int mypot[],hispot[],pointer[],turn; {
     ...
   }
4) void finish() { ... }                   //called when bait program ends.

The parameters mypot[c] and hispot[c] contains the pots at turn c.
pointer[c] contains the pointer at turn c, in [-3;3] where positive
(resp. negative) values means the notch is near us (resp. the opponent).
turn contains... err... the current turn. If your computer player doesn't
need initialisations, etc, you can use the noinit(),noendgame() and
nofinish() routines instead. ATM the only routine using initialisations
is maketable, so refer to its code if you want to see how to use all this.
Moreover maketable also creates and use an extra memory area.

Once the routines are written, you must append your player's data and
functions' pointers in the players' table (find "player[nbplayers]").
The structure of an entry can be seen below (find "typedef struct"), and
will look like:
     {"name","comment",init,bet,endgame,finish,},
Please note the extra pointer which is not initialised at the end of the
entry, and leave it as is. This is used for an extra memory pointer,
as you can see in maketable.

Most probably you will only need to write the bet routine, as is the case
in human, footsteps, triangle and dumb players, which are good starting
examples (esp. dumb IMHO).

That's it, now you can test your program against other players, and fill
the following result table which shows the win percentage when 100000
games are played (alas human player withdrew after 42 games! ;). Drawn
games are not taken in account.

  100000 games |   b: |   c: |   d: |   e: |   f: |   g: |
  -------------+------+------+------+------+------+------|
  b: footsteps |  --  |  17% |  44% |  36% |   0% |  47% |
  c: triangle  |  82% |  --  |  61% |  27% |   0% |   0% |
  d: dumb      |  55% |  38% |  --  |   7% |   0% |   0% |
  e: maketable |  63% |  72% |  92% |  --  |   ?  |   ?  |
  f: adapt     | 100% | 100% | 100% |   ?  |  --  | 100% |
  g: adapt2    |  52% | 100% | 100% |   ?  |   0% |  --  |

  Notes: maketable starts with an empty table against every opponent.
         Otherwise after "bait c e 10000" and "bait d e 10000", without
         removing the table, maketable wins 99% of his games against dumb.
*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
/* !!! #include <time.h>*/

#define nbplayers 7  /* Number of players' routines */
#define maxpot 50

/******************************************************** global variables */
int totalturns=0;
int p1,p2; /* Indices to players' structures */
FILE *filep,*fopen(const char *,const char *);
typedef struct {
  char *name;            /* name of the player */
  char *comment;         /* comment (author) */
  void (*initialise)();  /* pointer to initialisation function */
  int  (*bet)();         /* pointer to bet function */
  void (*endgame)();     /* pointer to endgame function */
  void (*finish)();      /* pointer to finish function */
  unsigned char *memory; /* pointer to extra memory if needed */
} playerstruct;

extern playerstruct player[];

/**** empty initialisation and finalization routines ***********************/
void noinit() {}
void noendgame(mypot,hispot,pointer,turn)
int mypot[],hispot[],pointer[],turn; {}
void nofinish() {}




/**** adapt2 by Alain Brobecker - 2004sep08 ********************************/
/* Same idea as adapt, but with randomness and better control.             */
#define AD2_N 10

/* Allocate memory for the number of turns in N games, and for a pointer   */
void AD2_init() {
  int c;
  player[6].memory=malloc(AD2_N+1);
  player[6].memory[0]=1; /* pointer to current game */
  for(c=1;c<=AD2_N;c++) {
    player[6].memory[c]=3;
  }
}

/* Compute average number of turns, and bet maxpot/average */
int AD2_bet(mypot,hispot,pointer,turn)
int mypot[],hispot[],pointer[],turn; {
  int c,total,turnsleft,bet;
  total=0;
  for(c=1;c<=AD2_N;c++) {
    total+=(int) player[6].memory[c];
  }
  turnsleft=(total/AD2_N)-turn; /* Nb of turns left = average - turn */
  if(turnsleft>0) {
    bet=(mypot[turn]/turnsleft); /* potleft/turnsleft */
    bet=bet-rand()%turnsleft; /* -1-rnd[-turnsleft;turnsleft] */
  } else {
    bet=1;
  }
  return (bet);
}

/* Enter the number of turn for current game in the memory */
void AD2_endgame(mypot,hispot,pointer,turn)
int mypot[],hispot[],pointer[],turn; {
  int ptr;
  ptr=player[6].memory[0];
  player[6].memory[ptr]=(unsigned char) turn;
  ptr++;
  if(ptr>AD2_N) { ptr=1; }
  player[6].memory[0]=(unsigned char) ptr;
}

/**** adapt by Alain Brobecker - 2004sep04 *********************************/
/* Computes the average number of turns over the last N games, then it     */
/* plays maxpot/average at each turn. Quite good though determinist.       */
#define AD_N 10

/* Allocate memory for the number of turns in N games, and for a pointer   */
void AD_init() {
  int c;
  player[5].memory=malloc(AD_N+1);
  player[5].memory[0]=1; /* pointer to current game */
  for(c=1;c<=AD_N;c++) {
    player[5].memory[c]=3;
  }
}

/* Compute average number of turns, and bet maxpot/average */
int AD_bet(mypot,hispot,pointer,turn)
int mypot[],hispot[],pointer[],turn; {
  int c,total,average;
  total=0;
  for(c=1;c<=AD_N;c++) {
    total+=(int) player[5].memory[c];
  }
  average=total/AD_N; /* >=1 since players must play >=1 turn to draw */
  return (maxpot/average);
}

/* Enter the number of turn for current game in the memory */
void AD_endgame(mypot,hispot,pointer,turn)
int mypot[],hispot[],pointer[],turn; {
  int ptr;
  ptr=player[5].memory[0];
  player[5].memory[ptr]=(unsigned char) turn;
  ptr++;
  if(ptr>AD_N) { ptr=1; }
  player[5].memory[0]=(unsigned char) ptr;
}

/**** maketable by Alain Brobecker - 2004aug29 *****************************/
/* !!! This one still needs further thinking. Basically it tries to find   */
/*     a winning play by attributing a coefficient for every configuration */
/*     and increasing the coefficient when it wins, decreasing it when it  */
/*     loses... */

/* To access table use ((mypot*(maxpot+1)+hispot)*5+pointer+2)*maxpot+bet  */
#define MT_memsize ((maxpot*(maxpot+1)+maxpot)*5+4)*maxpot+(maxpot-1)
#define MT_Max 255

/* Load coefficient table if it exists, else it is initialised to MT_Max/2 */
void MT_init() {
  int c;
  player[4].memory=malloc(MT_memsize);
  filep=fopen("table","rb");
  if(filep!=0) {
    fread(player[4].memory,1,MT_memsize,filep);
    fclose(filep);
  } else {
    for(c=0;c<MT_memsize;c++) {
      player[4].memory[c]=MT_Max/2;
    }
  }
}

/* We simply choose the bet (lower than mypot[turn]) that has the biggest  */
/* coefficient in this configuration (given mypot,hispot,pointer).         */
int MT_bet(mypot,hispot,pointer,turn)
int mypot[],hispot[],pointer[],turn; {
  int index,bestbet,c;
  unsigned char bestcoef;
  index=((mypot[turn]*(maxpot+1)+hispot[turn])*5+pointer[turn]+2)*maxpot;
  bestcoef=player[4].memory[index];
  bestbet=1;
  for(c=1;c<=mypot[turn];c++) {
    if(player[4].memory[index+c-1]>=bestcoef) { /* >= is better than > */
      bestcoef=player[4].memory[index+c-1];
      bestbet=c;
    }
  }
  return (bestbet);
}

/* Now we know the result of the game. If we won or the game was a draw,   */
/* we increment the coefficients for all choices made, else we decrement.  */
void MT_endgame(mypot,hispot,pointer,turn)
int mypot[],hispot[],pointer[],turn; {
  int index,c,delta,coef;
  if(pointer[turn]==3) {
    delta=1;
  } else if(pointer[turn]==-3) {
    delta=-1;
  } else {
    delta=0;
  }
  for(c=0;c<turn;c++) {
    index=((mypot[c]*(maxpot+1)+hispot[c])*5+pointer[c]+2)*maxpot;
    index+=mypot[c]-mypot[c+1]-1; /* the bet we chose -1 */
    coef=(int) player[4].memory[index]+delta;
    if( (coef>=0) && (coef<=MT_Max) ) {
      player[4].memory[index]=(unsigned char) coef;
    }
  }
}

/* save coefficients table */
void MT_finish() {
  filep=fopen("table","wb");
  if(filep!=0) {
    fwrite(player[4].memory,1,MT_memsize,filep);
    fclose(filep);
  }
}

/**** dumb by Alain Brobecker - 2004aug29 **********************************/
/* Since the average nb of turns seems to be around 8, i decided to make   */
/* this simple automaton which always bet around 50/8 ~ 6 tokens.          */
int dumb(mypot,hispot,pointer,turn)
int mypot[],hispot[],pointer[],turn; {
  return (4+(rand()%5)); /* 4+[0;4]*/
}

/**** triangle by Alain Brobecker - 2004aug28 ******************************/
/* I multiply max/2 (which depends upon how much the pointer is far from   */
/* the center position) by a random number, which probability rule is a    */
/* triangle (hence the name). */
int triangle(mypot,hispot,pointer,turn)
int mypot[],hispot[],pointer[],turn; {
  int bet;
  double r,max;
  bet=rand()%1001; /* bet in [0;1000] */
  r=((double) bet)/1000.0; /* r is in [0;1] */
  max=(double) mypot[turn]/(double) (4.0 - (double) abs(pointer[turn]));
  if(r>0.5) {
    bet=(int) (0.5+max*sqrt(r/2));
  } else if (r<0.5) {
    bet=(int) (0.5+max*(1-sqrt((1-r)/2)));
  } else {
    bet=(int) (0.5+max/2);
  }
  return (1+bet);
}

/**** footsteps by Gordon Munroe - 2000jun04 *******************************/
/* Litteral BASIC to C conversion of the game Gordon coded for CodeCraft#2 */
int footsteps(mypot,hispot,pointer,turn)
int mypot[],hispot[],pointer[],turn; {
  int bet;
  bet=0;
  if(mypot[turn]>0) {
    if(mypot[turn]>25) {
      bet=(rand()%18)+1; /* random value in [1;18] */
      bet=(bet%15)+1; /* now in [1;15], but [1;4] are more frequent */
      if(turn>0) {
        if(hispot[turn]-hispot[turn-1]==1) {
          bet=(rand()%10)+1;
        }
      }
    }
    if( (pointer[turn]==2) && (mypot[turn]>15) ) { bet=(rand()%5)+1+6; }
    if( (bet==0) && (mypot[turn]>10) ) { bet=(rand()%7)+1; }
    if( (bet==0) && (mypot[turn]>5) ) { bet=(rand()%4)+1; }
    if(hispot[turn]==0) { bet=1; }
    if( (pointer[turn]==1) && (mypot[turn]>hispot[turn]) && ((rand()%3)==0) ) {
      bet=hispot[turn]+1;
    }
    if(bet==0) { bet=1; }
  }
  return (bet);
}

/**** human player *********************************************************/
int human(mypot,hispot,pointer,turn)
int mypot[],hispot[],pointer[],turn; {
  int bet;
  printf(", your bet: ");
  scanf("%d",&bet);
  return (bet);
}

/************************************************ players' data & pointers */
/* Do not change the order in which they are set (esp 0;4).                */
playerstruct player[nbplayers] = {
  {"human","hey that's you!",noinit,human,noendgame,nofinish,},
  {"footsteps","by Gordon Munroe",noinit,footsteps,noendgame,nofinish,},
  {"triangle","by Alain Brobecker",noinit,triangle,noendgame,nofinish,},
  {"dumb","by Alain Brobecker",noinit,dumb,noendgame,nofinish,},
  {"maketable","by Alain Brobecker",MT_init,MT_bet,MT_endgame,MT_finish,},
  {"adapt","by Alain Brobecker",AD_init,AD_bet,AD_endgame,nofinish,},
  {"adapt2","by Alain Brobecker",AD2_init,AD2_bet,AD2_endgame,nofinish,} };

/***************************************************************** onegame */
/* returns 0 for draw, 1 when p1 wins and 2 when p2 wins. Set the print    */
/* variable to true when we must print informations on the game            */
int onegame(print) int print; {
  int p1pot[maxpot],p2pot[maxpot]; /* the pots during the game */
  int p1pointer[maxpot],p2pointer[maxpot]; /* both pointers are inversed */
  int turn,p1bet,p2bet,c,result;
  result=-1;
  turn=0;
  p1pot[0]=maxpot;
  p2pot[0]=maxpot;
  p1pointer[0]=0;
  p2pointer[0]=0;
  while(result==-1) {
    /* Print players' pots and the score */
    if(print) {
      printf("%s: %2d   ",player[p1].name,p1pot[turn]);
      for(c=3;c>p1pointer[turn];c--) { printf("."); }
      printf("o");
      for(c=3;c>p2pointer[turn];c--) { printf("."); }
      printf("   %s: %2d",player[p2].name,p2pot[turn]);
    }
    /* Ask the bets of both players */
    p1bet=player[p1].bet(p1pot,p2pot,p1pointer,turn);
    if( (p1bet<0) || (p1bet>p1pot[turn]) ) { p1bet=p1pot[turn]; }
    if( (p1pot[turn]>0) && (p1bet==0) ) { p1bet=1; }
    p1pot[turn+1]=p1pot[turn]-p1bet;
    p2bet=player[p2].bet(p2pot,p1pot,p2pointer,turn);
    if( (p2bet<0) || (p2bet>p2pot[turn]) ) { p2bet=p2pot[turn]; }
    if( (p1pot[turn]>0) && (p1bet==0) ) { p1bet=1; }
    p2pot[turn+1]=p2pot[turn]-p2bet;
    turn++;
    /* Print players' bets */
    if(print) {
      if( (p1!=0) && (p2!=0) ) { printf("\n"); } /* human pressed return */
      printf("%s: %2d   ",player[p1].name,p1bet);
      printf("  bet  ");
      printf("   %s: %2d\n",player[p2].name,p2bet);
    }
    /* Actualise pointers (and game?) according to bets */
    if(p1bet>p2bet) {
      p1pointer[turn]=p1pointer[turn-1]+1;
      if(p1pointer[turn]==3) { result=1; }
    } else if(p1bet<p2bet) {
      p1pointer[turn]=p1pointer[turn-1]-1;
      if(p1pointer[turn]==-3) { result=2; }
    } else if(p1bet==p2bet) {
      p1pointer[turn]=p1pointer[turn-1];
    }
    p2pointer[turn]=-p1pointer[turn];
    if((p1pot[turn]==0) && (p2pot[turn]==0)) { result=0; }
  }
  player[p1].endgame(p1pot,p2pot,p1pointer,turn);
  player[p2].endgame(p2pot,p1pot,p2pointer,turn);
  totalturns+=turn;
  return (result);
}

/******************************************************************* main */
int main(argc,argv) int argc; char *argv[]; {
  int c,win1=0,win2=0,nbgames,print;
  float avgturns;
  /* initialise random seed using cpu time */
/* !!!  time_t *temp;
   !!!  srand((time(temp))%9999); */
  /* Wrong number of arguments? */
  if( (argc<3) || (argc>4) ) {
    printf("\nUsage: 'bait a b' for one game between human and footsteps\n");
    printf("   or: 'bait b c N' for N games footsteps and triangle\n");
    printf("Possible players are:\n");
    for(c=0;c<nbplayers;c++) {
      printf("    %c:%s (%s)\n",((char)c)+'a',player[c].name,player[c].comment);
    }
    exit(1);
  }
  /* Get players' indices */
  p1=(int)(argv[1][0]-'a');
  player[p1].initialise();
  printf("%s (%s)  ",player[p1].name,player[p1].comment);
  p2=(int)(argv[2][0]-'a');
  player[p2].initialise();
  printf("vs  %s (%s)\n",player[p2].name,player[p2].comment);
  if((p1==0)&&(p2==0)) { printf("Humans can't play together\n"); exit(1); }
  /* if user gave 3 arguments then play only one game, if he */
  /* gave 4 arguments then read how many games to play */
  if(argc==3) {
    nbgames=1;
  } else {
    nbgames=0;
    c=0;
    while( (argv[3][c]>='0') && (argv[3][c]<='9') ) {
      nbgames=nbgames*10+(int) (argv[3][c]-'0');
      c++;
    }
    if(nbgames==0) { nbgames=1; }
  }
  /* print details when a player is human or nbgames==1 */
  print=(p1==0) || (p2==0) || (nbgames==1);
  /* go on */
  for(c=0;c<nbgames;c++) {
    switch(onegame(print)) {
      case 0:
        if(print) { printf("**** Draw game! ****\n"); }
        break;
      case 1:
        if(print) { printf("**** %s won! ****\n",player[p1].name); }
        win1++; break;
      case 2:
        if(print) { printf("**** %s won! ****\n",player[p2].name); }
        win2++; break;
    }
  }
  c=win1+win2;
  if(nbgames>1) {
    printf("  %s won %d games (%d%%)\n",player[p1].name,win1,100*win1/c);
    printf("  %s won %d games (%d%%)\n",player[p2].name,win2,100*win2/c);
    printf("  %d draw games\n",nbgames-c);
    avgturns=(float)totalturns/(float)nbgames;
    printf("Average number of turns: %f\n",avgturns);
  }
  player[p1].finish();
  player[p2].finish();
}
